import sys
sys.setrecursionlimit(10**2)
import random

from .ExpressionTree1 import ExpressionTree
from .FunctionalProcessingUnit import FunctionProcessingUnit as FPU
from .LoopProcessingUnit2 import LoopProcessingUnit
from .ConditonalProcessingUnit import ConditionalProcessingUnit as CPU
from .StringAlgorithm.ConditionSplit import  condition_split
from .StringAlgorithm.TokenizeVariableDefinition import tokenize_defining_variable
from .IOSTREAM.InputOutput import InputOutputStream as iostream
from .ProcessSetElements import process_logic_element
from .ProcessFunction import  process_functions
from .StringAlgorithm.ExtractFuncNameParams import extract_name_param
from .StringAlgorithm.SplitString import split
from .ClassProcessingUnit import accessing_attributes
from .ClassProcessingUnit import ClassProcessingUnit as CLSPU


class Compiler :
	
	def __init__(self,indent=4) :
		self.indent = indent
		self.loop_types = ['for','while']
		self.DataTypes = ['int','float','chars','let']
		self.ArrayDataTypes = ['int_arr','float_arr']
		self.int_literals = '0123456789'
		self.logic_elements = '∈∀∅∃∧∨∉'
		self.element = self.logic_elements[0]
		self.mathematical_constants = {'π' : 3.14159265,'ë' : 2.71828182}
		self.alpha = 'abcdefghijklmnopqrstuvwxyz'
		self.char_literals = 'abcdefghijklmnopqrstuvwxyz' + self.int_literals+'*/+-_^'
		self.variable_error = "Error variable {} is not defined "
		self.value_error = "Incorrect value {}assigned to the variable {}"
		self.break_statement  = 'over'
		self.continue_statement = 'skip'
		self.comment_syntax = '%'
		self.symbols = ["'",' ','.',">",'<','<=','=','/','?','*','-']
		self.iostream = iostream(self)
		self.func_body = {}
		self.functions = {}
		self.localVariables = {}
		self.func_scope = {}
		self.arithmetic_op = ['+','-','*','/','^']
		self.loop_scope = {}
		self.special_variables = '∆≠≥≤<'
		self.char_literals = self.char_literals + self.special_variables
		self.local_scope = {'func_scope' : self.func_scope,'loop_scope':self.loop_scope}
		self.localVariables = {'local_scope' : self.local_scope}
		self.globalVariables = {}
		self.Variables = {'global' : self.globalVariables,'local' : self.localVariables}
		self.literals = {'int' :self.int_literals,'float' : self.int_literals,'chars' : self.char_literals}
	
	def check_variable_name(self,name) :
		if name == '' :
			print("Error Name has not provided to the variable")
			return False
		if name[0].lower() in self.alpha +'_' :
			for i in range(1,len(name)) :
				if not name[i].lower() in self.alpha+self.int_literals+'_' :
					return False
			return True
		return False
	
	def has_indent(self,line,indent=0) :
		count = 0
		for l in line :
			if l == ' ' :
				count += 1
				continue
			break
		if count == self.indent+indent :
			return True
		return False
	
	def get_func_param(self,line) :
		if type(line) != type('a') :
			return 0,0,False
		name = ""
		param = ""
		ob_count = 0
		is_found = False
		N = len(line)
		
		for i in line :
			if i == '(' :
				is_found = True
				continue
			if is_found :
				if i == ')' :
					break
				param += i
				continue
			name += i
		
		return name,param,is_found
	
	def fetch_var_value(self,name,scope='global',scope_name='') :
		is_var = False
		if self.check_variable_name(name) :
				is_var = True
		if is_var :
			if scope == 'local' :
				var_scope = self.Variables[scope]['local_scope'][scope_name]
				if name in var_scope.keys() :
					val,_ = var_scope[name]
					return val
				if name in self.Variables['global'].keys() :
					val,_ = self.Variables['global'][name]
					return val
				#print("Error from fetch_var_value variable {} is not defined".format(name))
				return
			if scope == 'global' :
				if name in self.Variables[scope].keys() :
					val,_ = self.Variables[scope][name]
					return val
				#print("Error variable from fetch_var_value {} is not defined ".format(name))
				return
			return name
	
	def is_number(self,num) :
		is_num = True
		try :
			num = float(num)
			return True
		except :
			return False
		
		for n in num :
			if n == '.' :
				continue
			if not n in self.int_literals :
				is_num = False
				break
		return is_num
	
	def process_functions(self,line,scope='global',scope_name='',exe=False) :
			#func f(x) = 2*x+3
			if type(line) == type('a') :
				line = line.split()
			
			if len(line) == 1 and exe :
				#f(x)
				#f(3,x)
				if not self.is_function(line[0]) :
					return
				line = line[0]

				name,params,_ = extract_name_param(line)
				params = split(params,',')
				if not name in self.func_body.keys() :
					return
				params_value = []

				for param in params :
					if self.is_number(param) :
						params_value.append(param)
					if self.is_function(param) :
						params_value.append(self.execute_Functions(param,scope=scope,scope_name=scope_name))
					elif self.check_variable_name(param) :
						
						val,_ = self.fetch_variable(param,scope=scope,scope_name=scope_name)
						params_value.append(val)
				params_key = self.func_body[name]['params']
				
				for key,val in zip(params_key,params_value) :
					self.Variables['local']['local_scope']['func_'+name][key] = [val,'float']
				
				expression = self.func_body[name]['expression']
				
				tree = ExpressionTree(scope='local',scope_name='func_'+name,obj=self)
				tree.buildTree(expression)
				value = tree.computeTree(tree.head)
				return str(value)
				
			if len(line) == 4 :
				#func f(x) = 2*x
				if line[0] != 'func' :
					return
				name,params,_ = self.get_func_param(line[1])
				params = params.split(',')
				
				if not self.check_variable_name(name) :
					return
				
				self.func_body[name] = {}
				self.func_body[name]['params'] = params
				self.func_body[name]['expression'] = line[-1]
				
				self.Variables['local']['local_scope']['func_'+name] = {}
				
				return
				
			
	
	def process_function(self,line,scope='global',scope_name='',exe=False,obj=None) :
		#func f(x) = 2*3+1
		line = line.split()
		st = "Error incorrect Syntax {}"
		if line == [] :
			return
		if len(line) == 1 :
			#f(x)
			name,param,_ = self.get_func_param(line[0])
			if name in self.func_body.keys() :
				#f(x,y)
				#f(res,2)
				tree,dic,param_f = self.func_body[name]
				params = param.split(',')
				for key,val in zip(param_f,params) :
					if self.is_number(val) :
						dic[key] = val
						continue
					value = self.fetch_var_value(val,scope=scope,scope_name=scope_name)
					dic[key] = value
				dic_ = dic
				tree.param_values = dic
				val = str(tree.computeTree(tree.head))
				tree.param_values = dic_
				return val
					
		if exe :
			#f(x) = g(x)*2
			#f(x,y) = x*g(y)/2
			#f(x) = x*g(res)/2
			#g(X)
			if not obj :
				return
			if type(line) == type('a') :
				line =line.split()
			if len(line) != 1 :
				print('Error function is not given alone {}'.format(line))
				return
			name,param,is_found = self.get_func_param(line[0])
			
			if not name in self.func_body.keys() :
				print("Error function {} is not defined".format(name))
				return
			tree_,g_dict,g_params = self.func_body[name]
			g_dict_ = g_dict
			has_found_all = True
			to_find = []
			for key in param :	
				if key in obj.param_values.keys() :
					g_dict[key] = obj.param_values[key]
					continue
				else :
					to_find.append(key)
					has_found_all = False
			
			if not has_found_all :
				for key in to_find :
					g_dict[key] = self.fetch_var_value(key,scope=scope,scope_name=scope_name)
					has_found_all = True
			tree_.param_values = g_dict
			output = str(tree_.computeTree(tree_.head))
			tree_.param_values = g_dict_
			return output
			
		if len(line) == 4 :
			if line[0] != 'func' :
				#print(st.format(line))
				return
			func_syntax = line[1]
			name = ""
			param = ""
			is_found = False
			name,param,is_found = self.get_func_param(line[1])
			
			if not is_found :
				print(st.format(line))
				return
			
			if not self.check_variable_name(name) :
				print("Error incorrect function name {}".format(name))
				return
			self.param_dict = {}
			for i in param :
				if i == ',' :
					continue
				if not i in self.alpha :
					print("Error parameter must be alphabets {}".format(param))
					return
			
			params = param.split(',')
			for p in params :
				self.param_dict[p] = ''
			expr = line[3]
			
			if line[2] != '=' :
				print(st.format(line))
				return
			
			has_op = False
			
			for op in self.arithmetic_op :
				if op in expr :
					has_op = True
			if has_op :
				tree = ExpressionTree(self.Variables,is_func=True,parameters=params,obj=self)
				tree.buildTree(expr)
				self.func_body[name] = [tree,self.param_dict,params]
				return
			self.func_body[name] = expr
			return
		if len(line) == 3 :
			#y = f(x)
			
			name,_,_ = self.get_func_param(line[2])
			if not name in self.func_body.keys() :
				return
			
			for op in self.arithmetic_op :
			    if op in line[2] :
			        return
			var_y = line[0]
			var_scope = None
			if scope == 'local' :
				is_found = False
				if var_y in self.Variables['local']['local_scope'][scope_name].keys() :
					var_scope = self.Variables['local']['local_scope'][scope_name]
					is_found = True
				if not is_found :
					if var_y in self.Variables['global'].keys() :
						var_scope = self.Variables['global']
						is_found = True
				if not is_found :
					print("Error from process_func variable  {} is not defined".format(var_y))
					return	
			if scope == 'global' :
				if not var_y in self.Variables['global'].keys() :
					print("Error from process_unit variable  {} is not defined".format(var_y))
					return
				var_scope = self.Variables['global']
			name,param,is_found = self.get_func_param(line[2])
			if not is_found :
				#print(st.format(line))
				return
			if not name in self.func_body.keys() :
				print("Error function with name {} is not defined".format(name))
				return
			params = param.split(',')
			params_ = self.func_body[name][2]
			param_dict = self.func_body[name][1]
			for i,p in enumerate(params) :
				is_var = False
				for e in p :
					if e in self.alpha :
						is_var = True
						break
				if is_var :
					if scope == 'local' :
						if p in self.Variables['local']['local_scope'][scope_name].keys() :
							v,_ = self.Variables['local']['local_scope'][scope_name][p]
							param_dict[params_[i]] = v
							continue
						if p in self.Variables['global'].keys() :
							v,_= self.Variables['global'][p]
							param_dict[params_[i]] = v
							continue
						print("Error variable from process_func unit {} is not defined".format(p))
						return
					if scope == 'global' :
						if p in self.Variables['global'].keys() :
							v,_= self.Variables['global'][p]
							param_dict[params_[i]] = v
							continue
						#print("Error variable {} is not defined".format(p))
						return
				param_dict[params_[i]] = params[i]
				
			tree = self.func_body[name][0]
			tree.param_values = param_dict
			value = str(tree.computeTree(tree.head))
			_,dt = var_scope[var_y]
			var_scope[var_y] = [value,dt]
			return True
			
	def process_Func(self,line,scope='global',scope_name='') :
			#y = pow(x)
			#where pow is a multiline function
			line = line.split()
			
			
			if len(line) == 1 :
				#pow(x,r)
				if self.is_function(line[0]) :
					name,_,_ = self.get_func_param(line[0])
					if name in self.functions.keys() :
						func = self.functions[name]
						if scope_name :
							func.outer_scope = scope_name
						return func.executeTree(line[0])
				return
			
			# y = 2*pow(x)
			if len(line) != 3 :
				return
				
			if not self.get_func_param(line[-1]) :
				return
			
			func_name = line[-1]
			name,param,_ = self.get_func_param(func_name)
			func = None
			if name in self.functions.keys() :
				func = self.functions[name]
				val = None
			else :
				return

			if len(condition_split(line[-1],'any')) > 1 :

					val = str(self.compute_index(line[-1],scope=scope,scope_name=scope_name,type='float'))
					if self.get_arr_index(line[0]) :
						self.process_array(line,scope=scope,scope_name=scope_name)
						return
					self.assign_value(line[0],val,scope=scope,scope_name=scope_name)
					return
					
			func.outer_scope = scope_name
			val = func.executeTree(line)
			var_y = line[0]
			
			if self.get_arr_index(var_y) :
					#arr[index] = func(y)
				name,index,_ = self.get_arr_index(var_y)
				if len(index.split(',')) == 2 :
					#this means var_y is an 2-d array
					row,col = self.compute_coords(index.split(','),scope=scope,scope_name=scope_name)
					self.assign_2darray(name,val,x=row,y=col,scope=scope,scope_name=scope_name)
					return
				if len(condition_split(index,'any')) > 1 :
					index = self.compute_index(index,scope=scope,scope_name=scope_name)
				elif self.check_variable_name(index) :
					index = self.fetch_var_value(index,scope=scope,scope_name=scope_name)
					index = int(index)
					self.assign_array(name,val,index=index,scope=scope,scope_name=scope_name)
					return
				
					
			self.assign_value(var_y,val,scope=scope,scope_name=scope_name)
			return
			"""
			if scope == 'global' :
				if var_y in self.Variables['global'].keys() :
					_,dt = self.Variables['global'][var_y]
					val = self.format_literals(dt,str(val))
					self.Variables['global'][var_y] = [val,dt]
					return
				
			if scope == 'local' :
				if scope_name :
					if var_y in self.Variables['local']['local_scope'][scope_name].keys() :
						self.assign_value(var_y,val,scope=scope,scope_name=scope_name)
						return	
			
			
			if var_y in self.Variables['local']['local_scope'][name].keys() :
				_,dt = self.Variables['local']['local_scope'][name][var_y]
				val  = self.format_literals(dt,str(val))
				self.Variables['local']['local_scope'][name][var_y] = [val,dt]
				return
			if var_y in self.Variables['global'].keys() :
				_,dt = self.Variables['global'][var_y]
				val  = self.format_literals(dt,str(val))
				self.Variables['global'][var_y] = [val,dt]
				return
			else :
				return
			"""
			
	def clean_line_from_op(self,line) :
		clean_line = ''
		for l in line :
			if l in self.arithmetic_op :
				clean_line += " "
				continue
			clean_line += l
		return clean_line
	
	def get_arr_index(self,arr) :
		if self.is_a_number(arr) :
			return False
		if len(arr) == 0 :
			return False
			
		name = ""
		index = ""
		count = 0
		is_arr = False
		found_brack = False
		N = len(arr)
		if arr[-1] != ']' :
			return False
		count = 1
		for i in range(N-1) :
			if arr[i] == '[' :
				count += 1
				found_brack = True
				continue
			if not found_brack :
				name += arr[i]
				continue
			index += arr[i]
		
		if count == 2 :
			is_arr = True
		if not self.check_variable_name(name) :
			is_arr = False
		return name,index,is_arr
		
	def have_two_brackets(self,var) :
		count = 0
		for i in var :
			if i == '[' or i == ']' :
				count += 1
		if count == 2 :
			return True
		return False
	
	def is_char(self,value) :
		if len(value) < 2 :
			return False
		if value[0] == "'" and value[-1] == "'" :
			return True
		return False
	
	def is_var_exist(self,var,scope='global',scope_name='') :
		if scope == 'global' :
			if var in self.Variables[scope].keys() :
				return True
			return False
		if var in self.Variables[scope]['local_scope'][scope_name].keys() :
			return True
		return self.is_var_exist(var,scope='global')
	
	def is_variable_in_scope(self,var,scope='global',scope_name='') :
		if scope == 'local' :
			if var in self.Variables[scope]['local_scope'][scope_name].keys() :
				return True
			return False
		if var in self.Variables['global'].keys() :
			return True
		return False
	
	def fetch_variable(self,var,scope='global',scope_name='') :
		if scope == 'global' :
			return self.Variables[scope][var]
		if var in self.Variables['local']['local_scope'][scope_name].keys() :
			data,dt = self.Variables['local']['local_scope'][scope_name][var]
			if data != None :
				return data,dt
		return self.fetch_variable(var,scope='global')
		
	def assign_array(self,name,value,index=None,scope='global',scope_name="") :
		if index != None :
			if scope == 'local' :
				if name in self.Variables['local']['local_scope'][scope_name].keys() :
					elements,dt = self.Variables['local']['local_scope'][scope_name][name]
					elements[index] = self.format_literals(dt[:3],value)
					self.Variables['local']['local_scope'][scope_name][name] = [elements,dt]
					return
			if name in self.Variables['global'].keys() :
				elements,dt = self.fetch_variable(name,scope='global')
				elements[index] = self.format_literals(dt[:3],value)
				self.Variables['global'][name] = [elements,dt]
				return
		if name in self.Variables['global'].keys() :
			elements,dt = self.fetch_variable(name,scope='global')
			elements[index] = self.format_literals(dt[:3],value)
			self.Variables['global'][name] = [elements,dt]
			return
		return self.assign_value(name,value,scope=scope,scope_name=scope_name)
		
	def define_variable(self,variable,value,data_type,scope='global',scope_name="") :
		if scope == 'global' :
			self.Variables[scope][variable] = [value,data_type]
			return
		if scope == 'local' :
			self.Variables[scope]['local_scope'][scope_name][variable] = [value,data_type]
			return
		
	def assign_value(self,var,val,scope='global',scope_name='') :
		if scope == 'global' :
			if not self.is_var_exist(var) :
				print("Variable {} is not defined ".format(var))
				return
			_,dt = self.Variables[scope][var]
			if type(val) == type('a') :
				val = self.format_literals(dt,val)
			self.Variables[scope][var] = [val,dt]
			return
		if var in self.Variables[scope]['local_scope'][scope_name].keys() :
			_,dt  = self.Variables[scope]['local_scope'][scope_name][var]
			if type(val) == type('a') :
				val = self.format_literals(dt,val)
			self.Variables[scope]['local_scope'][scope_name][var] = [val,dt]
			return
		return self.assign_value(var,val,scope='global')
		
	def compute_index(self,index,scope='global',scope_name='',type='int') :
		if index :
			if self.is_number(index) :
				return int(index)
			has_op = False
			val = condition_split(index,'any')
			if len(val) > 1 :
				has_op = True
			del val
			if has_op :
				tree = ExpressionTree(Variables=self.Variables,scope=scope,scope_name=scope_name,obj=self)
				tree.buildTree(index)
				return_value = tree.computeTree(tree.head)
				if type == 'int' :
					return int(return_value)
				if type == 'float' :
					return float(return_value)
			return_value = self.fetch_var_value(index,scope=scope,scope_name=scope_name)
			if type == 'int' :
				return int(return_value)
			if type == 'float' :
				return float(return_value)
			if type == 'let' :
				return float(return_value)
			
	def get_arr_element(self,name,index,scope='global',scope_name="") :
		if type(index) == type('a') :
			index = self.compute_index(index,scope=scope,scope_name=scope_name)
		elements,_ = self.fetch_variable(name,scope=scope,scope_name=scope_name)
		if index < len(elements) :
			return elements[index]
		print("index {} out of range for array of size {}".format(index,len(elements)))
		
	def has_arithmetic_op(self,line) :
		for op in self.arithmetic_op :
			if op in line :
				return True
		return False
		
		
	def is_2darray(self,name) :
		count_brackets = 0
		index = ''
		found_bracket = False
		
		for char in name :
			if char == '[' :
				found_bracket = True
				continue
			if char == ']' :
				break
			if found_bracket :
				index += char
		if len(index.split(',')) == 2 :
			return True
		return False
		
	def get_2darray(self,line) :
		#arr[i+1][j-1]
		row = ""
		column = ""
		name = ""
		count_b = 0
		
		for char in line :
			have_found_b = False
			if char == '[' :
				have_found_b = True
				count = count + 1
				continue
			if char == ']' :
				have_found_b = False
				continue
			if have_found_b :
				if count == 1 :
					row += char
				else :
					column += char
			name += char
		
		return name,row,column
	
			
	def fetch_2darray_element(self,value,data_type) :
		row_elements = []
		column_elements = []
		N = len(value)
		count_b = 0
		for i,v in enumerate(value) :
			if v == '(' :
				column_elements  = []
				count_b = 1
				continue
			if v == ')' :
				if i+1 >= N :
					break
				row_elements.append(column_elements)
				count_b = 0
				continue
			if count_b :
				if v == ',' :
					continue
				v =self.format_literals(data_type[:3],v)
				column_elements.append(v)
				
		return row_elements
		
	def compute_coords(self,coords : list,scope='global',scope_name='',another_scope='') :
		if coords == [] :
			return
		row,column = coords
		compute_row = compute_col = False
		
		if len(condition_split(row,'any')) > 1 :
			row = self.compute_index(row,scope=scope,scope_name=scope_name)
			compute_row = True
		if len(condition_split(column,'any')) > 1 :
			column = self.compute_index(column,scope=scope,scope_name=scope_name)
			compute_col  = True
		
		if not compute_row :
			if self.is_number(row) :
				row = int(float(row))
			elif self.check_variable_name(row) :
				row = self.fetch_var_value(row,scope=scope,scope_name=scope_name)
				row = int(float(row))
		
		if not compute_col :
			if self.is_number(column) :
				column = int(float(column))
			elif self.check_variable_name(column) :
				column = self.fetch_var_value(column,scope=scope,scope_name=scope_name)
				column = int(float(column))
		
		return row,column
		
	def assign_2darray(self,name,val,x=None,y=None,scope='global',scope_name='') :
		
		elements,dt = self.fetch_variable(name,scope=scope,scope_name=scope_name)
		elements[x][y] = self.format_literals(dt[:3],val)
		
		if scope == 'local' :
			if name in self.Variables['local']['local_scope'][scope_name].keys() :
				self.Variables['local']['local_scope'][scope_name][name] = [elements,dt]
				return
			if name in self.Variables['global'].keys() :
				self.Variables['global'][name] = [elements,dt]
				return
		
		if name in self.Variables['global'].keys() :
			self.Variables['global'][name] = [elements,dt]
		return
			
	
	def fetch_2darray(self,name,scope='global',scope_name='') :
		if scope == 'global' :
			if name in self.Variables['global'].keys() :
				return self.Variables['global'][name]
		if name in self.Variables['local']['local_scope'][scope_name].keys() :
			return self.Variables['local']['local_scope'][scope_name][name]
		return self.fetch_2darray(name,scope='global')
		
	def process_array(self,line,scope='global',scope_name='',class_obj=None,another_scope='') :
		if type(line) == type('a') :
			line = line.split()
		#int_arr arr[] = (1,2,3,4,5,6)
		#arr[6] = 7
		#arr = (1,2,3,4,5)
		#int2d_arr array[][] = ((1,2,3,4,5),(6,7,8,9,10))
		
		if '∈' in line :
			return
		
		if len(line) == 1 :
			#arr[i]
			# i E W
			
			if self.is_2darray(line[0]) :
				#2d array
				#arr[i,j]
				if not self.get_arr_index(line[0]) :
					return
				name,coords,_ = self.get_arr_index(line[0])
				row,column = coords.split(',')
				row,column  = self.compute_coords((row,column),scope=scope,scope_name=scope_name,another_scope=another_scope)
				elements,dt = self.fetch_2darray(name,scope=scope,scope_name=scope_name)
				m,n = len(elements),len(self.to_int(elements[0]))
				
				if row > m or column > n:
					print('IndexOutofBoundError [{},{}]'.format(m,n))
					return
				
				return elements[row][column],dt
						
			
			if not self.get_arr_index(line[0]) :
				return
			name,index,_ = self.get_arr_index(line[0])
			elements = []
			data_type = None
			line_op = condition_split(index,'any')
			has_op = False
			if len(line_op) > 1:
				index = self.compute_index(index,scope=scope,scope_name=scope_name)
				has_op = True
			elements,data_type = self.fetch_variable(name,scope=scope,scope_name=scope_name)
			if not has_op :
				if not index in self.int_literals :
					if self.check_variable_name(index) :
						index = self.fetch_var_value(index,scope=scope,scope_name=scope_name)
			index = int(index)
			if index < len(elements) :
				return elements[index],data_type
			print("error index {} out of range in array {}".format(index,line[0]))
			return
			
		
		if len(line) == 4 :
			#int_arr arr[] = (1,2,3,4,5,6,7)
			#float_arr arr[] = (1.23,9.9)
			#int_arr arr[5+1] = (1,2,3,4,5)
			#int row = arr[1]

			if line[0] in self.DataTypes :
				value = None
				new_line = None
				if len(condition_split(line[3],'any')) > 1 :
					try :
						value = self.compute_index(line[3],scope=scope,scope_name=scope_name,type='float')
						new_line = line[0] + ' ' + line[1] + ' ' + line[2]+ ' '+str(value)
						self.process_variable(new_line,scope=scope,scope_name=scope_name)
					except :
						pass
						
				elif self.get_arr_index(line[3]) :
					value,_ = self.process_array(line[3],scope=scope,scope_name=scope_name)
					new_line = line[0] +'  '+ line[1] +'  '+ line[2]+'  ' + str(value)
					self.process_variable(new_line,scope=scope,scope_name=scope_name)
				return
			
			if line[0] in self.ArrayDataTypes :
				data_type = line[0]
				name = line[1]
				
				if self.is_2darray(name) :
					#int_arr arr[2,7]  = ((1,2,3,4,5,6,7),(8,9,10,11,12,13,14))
					name,coords,_ = self.get_arr_index(name)
					coords = coords.split(',')
					self.check_variable_name(name)
					value = line[3]
					elements = None
					if self.check_variable_name(value) :
						elements = self.fetch_var_value(value,scope=scope,scope_name=scope_name)
					else :
						elements = self.fetch_2darray_element(value,data_type)
					m,n = len(elements),len(elements[0])
					row,column = self.compute_coords(coords,scope=scope,scope_name=scope_name)
					
					if row > m or column > n :
						print('error the index entered at the time of initialising the array is wrong')
						return
					
					self.define_variable(name,elements,data_type,scope=scope,scope_name=scope_name)
					if class_obj :
						class_obj.head.body[name] = elements
					return True
				
				is_arr= self.get_arr_index(name)
				if not is_arr :
					return
				name,index,_ = self.get_arr_index(name)
				index = self.compute_index(index,scope=scope,scope_name=scope_name)
				temp_elements = ''
				count = 0
				if self.check_variable_name(line[-1]) :
					line[-1] = self.fetch_var_value(line[-1],scope=scope,scope_name=scope_name)
				for e in line[-1] :
					if e == '('  :
						count += 1
						continue
					if e == ')' :
						count += 1
						continue
					temp_elements += e 
				elements_delta = []
				for e in temp_elements.split(',') :
					if self.is_number(e) :
						elements_delta.append(e)
						continue
					elements_delta.append(self.fetch_var_value(e))
				#fetched values of the element	
				if not count >= 2 :
					print("Syntax error from process_array {}".format(line))
					return
				if index :
					if index != len(elements_delta) :
						print("index {} does not match the number of elements in the array {}".format(index,elements_delta))
						return
				if scope == 'local' :
					self.Variables['local']['local_scope'][scope_name][name] = [elements_delta,data_type]
					if class_obj :
						class_obj.head.body[name] = elements_delta
					return True
				self.Variables['global'][name] = [elements_delta,data_type]
				return True
		
		if len(line) == 3 :
			# y = arr[i]
			#arr[i] = y
			# t = arr[]
			#res = 2+arr[i]
			#y = arr[i,j]
			#arr[i] = f(x)
			y = line[0]
			arr = line[2]
			clean_line = condition_split(arr,'any')
			have_arr = False
			
			for element in clean_line+[y] :
				if self.get_arr_index(element) :
					have_arr = True
					break				
			
			if not have_arr :
				return

			has_op = True
			
			if len(clean_line) == 1 :
				has_op = False
				
			if has_op :
				#var = arr[i+k]*arr[h+k]
				#var = var*arr[i-k]
				is_2d = False
				row  = column = None
				tree = ExpressionTree(Variables=self.Variables,scope=scope,scope_name=scope_name,obj=self)
				tree.buildTree(arr)
				val  = str(tree.computeTree(tree.head))
				#if y is an array then this will be executed
				if self.get_arr_index(y) :
					
					name,index_y,_ = self.get_arr_index(y)

					if len(index_y.split(',')) > 1 :
						is_2d = True
						row,column = index_y.split(',')
						row,column = self.compute_coords((row,column),scope=scope,scope_name=scope_name)
						#arr[i,j] = arr[k+1,h+1]
						self.assign_2darray(name,val,x=row,y=column,scope=scope,scope_name=scope_name)
						return
					
					else :	
						return_value = condition_split(index_y,'any')
					
					if len(return_value) > 1 :
						index_y = self.compute_index(index_y,scope=scope,scope_name=scope_name)
					
					elif self.is_number(str(index_y)) :
						pass
					
					else :
						index_y = self.fetch_var_value(index_y,scope=scope,scope_name=scope_name)
					index_y = int(index_y)
					self.assign_array(name,val,index=index_y,scope=scope,scope_name=scope_name)
					return True
				#if y is a variable then this line will be executed
				self.assign_value(y,val,scope=scope,scope_name=scope_name)
				return True
			
			if  self.get_arr_index(arr) or  self.get_arr_index(y) :
				pass
			else :
				return
			
			#arr[i] = y
			
			
			if self.get_arr_index(y)  :
				
				value_found = False
				name,index,_ = self.get_arr_index(y)
				val = None
				is_2d = False
				if self.is_2darray(y) :
					is_2d = True
				row=column = None
				
				if self.get_arr_index(arr) :
					#arr[i,j] = arr[j,k]
					#arr[i,j] = arr[k]
					val,_ = self.process_array(arr,scope=scope,scope_name=scope_name)
					
				if self.is_2darray(y) :
					index = index.split(',')
					row,column = self.compute_coords(index,scope=scope,scope_name=scope_name)

				else :
					index = self.compute_index(index,scope=scope,scope_name=scope_name)
					
				elements,dt  = self.fetch_variable(name,scope=scope,scope_name=scope_name)

				if self.is_number(arr) :
					if is_2d :
						self.assign_2darray(name,arr,x=row,y=column,scope=scope,scope_name=scope_name)
						return True
					self.assign_array(name,arr,index=index,scope=scope,scope_name=scope_name)
					return True
				
				if self.check_variable_name(arr) :
					arr,_ = self.fetch_variable(arr,scope=scope,scope_name=scope_name)
					value_found = True
					val = arr
				
				if is_2d :
					if value_found :
						self.assign_2darray(name,arr,x=row,y=column,scope=scope,scope_name=scope_name)

				elif self.is_function(arr) :
					name_,params,_ = self.get_func_param(arr)
					val = None
					if name_ in self.func_body.keys() :
						val = self.process_functions(arr,scope=scope,scope_name=scope_name,exe=True)
					elif name_ in self.functions.keys() :
						func = self.functions[name]
						func.outer_scope = scope_name
						val = func.executeTree(arr)
					
					if is_2d :
						self.assign_2darray(name,val,scope=scope,x=row,y=column,scope_name=scope_name)
						return True
					self.assign_array(name,val,index=index,scope=scope,scope_name=scope_name)
					return True
					
				elif self.get_arr_index(arr) :
					
					#arr[i] = arr[j]
					#arr[i,j] = arr[k,l]
					#arr[i,k] = arr[j]
					#arr[k] = arr[k,l]
					name_j,j_index,_= self.get_arr_index(arr)
					row = column = None
					
					if len(j_index.split(',')) > 1 :
						value = self.process_array(arr,scope=scope,scope_name=scope_name)
						#in case if y is a 2 dimensional array
						if self.is_2darray(y) :
							row,column = self.compute_coords(index.split(','),scope=scope,scope_name=scope_name)
							self.assign_2darray(name,value,x=row,y=column,scope=scope,scope_name=scope_name)
							return
						self.assign_array(name,value,index=index,scope=scope,scope_name=scope_name)
						return
							
					else :
						j_index = self.compute_index(j_index,scope=scope,scope_name=scope_name)
						j_arr = self.get_arr_element(name_j,j_index,scope=scope,scope_name=scope_name)
						
						if self.is_2darray(y) :
							row,column = self.compute_coords(index.split(','),scope=scope,scope_name=scope_name)
							self.assign_2darray(name,j_arr,x=row,y=column,scope=scope,scope_name=scope_name)
							return
						
						elements[index] = j_arr
						
					if scope == 'local' :
						self.Variables[scope]['local_scope'][scope_name][name] = [elements,dt]
					else :
						self.Variables[scope][name] = [elements,dt]
					#self.assign_array(name,j_arr,index=index,scope=scope,scope_name=scope_name)
					return
					

				if not value_found :
					return
				
				if self.is_2darray(y) :
					if type(index) == type('a') :
						index = index.split(',')
					row,column = self.compute_coords(index,scope=scope,scope_name=scope_name)
					self.assign_2darray(name,arr,x=row,y=column,scope=scope,scope_name=scope_name)
					return
				
				self.assign_array(name,arr,index=index,scope=scope,scope_name=scope_name)
				return
			#y = arr[i]
			return_val = self.process_array(arr,scope=scope,scope_name=scope_name)
			
			if not return_val :
				return
			
			data_type = return_val[1].split('_')[0]
			self.assign_value(y,return_val[0],scope=scope,scope_name=scope_name)
			return
			
	def is_array(self,line) :
		for l in line :
			if l.split() == [] :
				continue
			if self.get_arr_index(l.split()[0]) :
				return True
		return False
		
	def is_a_number(self,n) :
		try :
			float(n)
			return True
		except :
			return False
		
	def is_function(self,line) :
		if self.is_a_number(line) :
			return False
			
		
		if not '=' in line :
			name,_,_ = extract_name_param(line)
			if name in self.functions.keys() :
				return True
			if name in self.func_body.keys() :
				return True
			return False
			
			if len(line) == 2 :
				return True
			return False
		if type(line) == type('a') :
			line = line.split('=')
		line  = line[1]
		# a = pow(2,3)
		if line.split('(')[0] in self.functions :
			return True
		if line.split('(')[0] in self.func_body :
			return True
		return False
		
	def check_number_literals(self,var,value) :
		for v in value :
			if v == '.' :
				continue
			if v == '-' :
				continue
			if not v in self.int_literals :
				return False
		return True
		
	def check_char_literals(self,var,value) :
		if value[0] != "'" or value[-1] != "'"  :
			return
		for v in value :
			if v in self.symbols :
				continue
			if not v.lower() in self.char_literals + self.int_literals :
				return False
		return True
		
	def format_char_literals(self,chars) :
		if chars == '' :
			return chars
		line = ''
		for i,char in enumerate(chars) :
			if i + 1 == len(chars) :
				line += char
				break
			line += char
			line += ' '
		return line
		
	def has_negative_sign(self,num) :
		if self.is_a_number(num) :
			num  = str(num)
		if num[0] == '-' :
			for op in ['*','/','+','^'] :
				if op in num :
					return False
			return True
		return False
	
	def format_negative_number(self,num) :
		if self.is_a_number(num) :
			num = str(num)
		if num[0] == '-' :
			return num[1:]
		return num
		
	def minus_minus_plus(self,num) :
		if len(num) >= 3 :
			if num[0] == '-' and num[1] == '-' :
				return num[2:]
		return num
	
	def print_scope_body(self,scope='local',scope_name='') :
		if scope == 'global' :
			print(self.Variables['global'])
			return
		if scope_name == '' :
			return
		print(self.Variables[scope]['local_scope'][scope_name])
	
	def format_literals(self,data_type,num) :
		if data_type == 'int' :
			
			if type(num) == type([1]) :
				for i,row in enumerate(num) :
					if type(row) == type([1]) :
						for j,col in enumerate(row) :
							if '.' in col :
								col = str(col)
								num[i][j] = col.split('.')[0]
					else :
						if '.' in row :
							num[i] = row.split('.')[0]
				return num
			num = str(num)
			if '.' in num :
				return num.split('.')[0]
			return num
		
		if data_type in ['float','let'] :
			if type(num) == type([1]) :
				for i,row in enumerate(num) :
					if type(row) == type([1]) :
						for j ,col in enumerate(row) :
							col = str(col)
							if not '.' in col :
								num[i][j] = col+'.0'
					else :
						if not '.' in row :
							num[i] = row +'.0'
				return num
			num = str(num)		
			if not '.' in num :
				num = num + '.0'
			return num
		return num
		
		
		
	def process_variable(self,line,scope='global',scope_name='',class_obj=None) :
		if self.is_array(line.split('=')) :
			return
		if self.is_function(line) :
			return
		if '∈' in line :
			return
		save = line
		line = line.split()
		if len(line) <= 2 :
			return
		N = len(line)
		#DataTypes Var_name = Value
		#Var_name = value
		if line[0] == 'func':
		    return
		
		if line[0] in self.DataTypes :
			N = 4
		else :
			N = 3
		
		if N == 4 :
			data_type = line[0]
			if data_type in self.ArrayDataTypes :
				return
			name = line[1]
			value = line[3]
			if line[2] != '=' :
				print("Syntax Error from process_variable unit {}".format(line))
				return
			if not self.check_variable_name(name) :
				return
					
			if data_type in ['int','float','let'] :
				has_operator = False
				is_negative = False
				
				val = condition_split(value,'any')
				if len(val) > 1 :
					has_operator = True
				del val
				
				if has_operator :
					value = str(self.compute_index(value,scope=scope,scope_name=scope_name,type='float'))
				#DataType Variable_a = variable_b
				#int a = b  
				#int a = f(x)
				#int element = arr[1]
				
				elif self.is_function(value) :
					name_,params,_ = self.get_func_param(value)
					
					if name_ in self.func_body.keys() :
						value = self.process_functions(value,scope=scope,scope_name=scope_name,exe=True)
					elif name_ in self.Variables['local']['local_scope'].keys() :
						func = self.functions[name_]
						func.outer_scope = scope_name
						value = func.executeTree(value)
				
				elif self.get_arr_index(value) :
					value,_ = self.process_array(value,scope=scope,scope_name=scope_name)
					value = str(value)
				else :
					is_negative = False
					
					if self.has_negative_sign(value) :
						value = self.format_negative_number(value)
						is_negative = True
						
					if self.check_variable_name(value) :
						value = self.fetch_var_value(value,scope=scope,scope_name=scope_name)
	
					if value in self.mathematical_constants.keys() :
						value = str(self.mathematical_constants[value])
						
					if is_negative :
						value = '-' + value
						value = self.minus_minus_plus(value)
						
				
				value = self.format_literals(data_type,value)
				self.define_variable(name,value,data_type,scope=scope,scope_name=scope_name)
				if class_obj :
					class_obj.head.body[name] = value
				return True
				
			if data_type == 'chars' :
				value = self.format_char_literals(line[3:])
				if not self.check_char_literals(name,value) :
					return
				self.define_variable(name,value,data_type,scope=scope,scope_name=scope_name)
				if class_obj :
					class_obj.head.body[name] = value
				return True
				
		if N == 3 :
			# a = 3
			if len(line) >= 4 :
				if line[0] in self.ArrayDataTypes or line[0] in self.DataTypes :
					return
			
			if scope == 'local' :
				if scope_name == '' :
					 print("Error from process_variable unit please provide the name of the scope ")
					 return
			
				
				name = line[0]
				value = line[2]
				
				if value in self.mathematical_constants.keys() :
					value = str(self.mathematical_constants[value])
					
				val = condition_split(value,'any')
				if len(val) > 1 :
					have_array = False
					for each_ele in val :
						if self.get_arr_index(each_ele) :
							return
				else :
					if self.get_arr_index(val[0]) :
						return
				if name in self.Variables[scope]['local_scope'][scope_name].keys() :
				   _,data_type = self.Variables[scope]['local_scope'][scope_name][name]
				   
				   if data_type in self.DataTypes :
				   	has_operator = False
				   	val = condition_split(value,'any')
				   	if len(val) > 1 :
				   		has_operator = True

				   	if has_operator :
				   	    value = str(self.compute_index(value,scope=scope,scope_name=scope_name,type='float'))
				   	    
				   	
				   	if self.is_function(value) :
				   		name_,params,_ = self.get_func_param(value)
				   		if name_ in self.func_body.keys() :
				   			value = self.process_functions(value,scope=scope,scope_name=scope_name,exe=True)
				   			value = str(value)
				   	
				   	if not self.check_number_literals(name,value) :
				   		return
				   	value = self.format_literals(data_type,value)
				   	self.assign_value(name,value,scope=scope,scope_name=scope_name)
				   	return True
				   if data_type == 'chars' :
				       value = self.format_char_literals(line[2:])
				       if not self.check_char_literals(name,value) :
				       	return
				       self.assign_value(name,value,scope=scope,scope_name=scope_name)
				       return True
				else :
					 #line = self.format_char_literals(line)
					# return self.process_variable(line,scope='global'))
					 #res = res*i
					 if name in self.Variables['global'].keys() :
					 	_,data_type = self.Variables['global'][name]
					 	
					 	if data_type in ['int','float','let'] :
					 		has_op = False
					 		val = condition_split(value,'any')
					 		
					 		if len(val) > 1 :
					 			has_op = True
					 				
					 		if has_op :
					 			value = str(self.compute_index(value,scope=scope,scope_name=scope_name,type='float'))
					 			
					 		if self.is_function(value) :
					 			name_,params,_ = self.get_func_param(value)
					 			if name_ in self.func_body.keys() :
					 				value = self.process_functions(value,scope=scope,scope_name=scope_name,exe=True)
					 				value = str(value)
					 				
					 		if not self.check_number_literals(name,value) :
					 			return
					 		value = self.format_literals(data_type,value)
					 		self.assign_value(name,value,scope=scope,scope_name=scope_name)
					 		return True
					 	if data_type == 'chars' :
					 		value = self.format_char_literals(line[2:])
					 		if not self.check_char_literals(name,value) :
					 			return
					 		self.assign_value(name,value,scope=scope,scope_name=scope_name)
					 		return True	
					 print("Error from process_variable unit assigning value before defining a variable {}".format(line))
					 return
			if scope == 'global' :
				if line[0] in self.ArrayDataTypes :
					return
				if self.is_function(save) :
					return

				name = line[0]
				value = line[2]
				
				if value in self.mathematical_constants.keys() :
					value = str(self.mathematical_constants[value])
					
				
				if line[1] != '=' :
					print("Syntax  Error from process_variable {}".format(line))
					return	
				if name in self.Variables[scope].keys() :
					_,data_type = self.Variables[scope][name]
					if data_type in self.DataTypes :
						has_operator = False
						val = condition_split(value,'any')
						if len(val) > 1 :
							has_operator = True
						del val
						if has_operator :
							value = str(self.compute_index(value,scope='global',type='float'))
						
						if self.is_function(value) :
							name_ ,params,_ = self.get_func_param(value)
							
							if name_ in self.func_body.keys() :
								value = self.process_functions(value,scope=scope,scope_name=scope_name,exe=True)
								value = str(value)

						if not self.check_number_literals(name,value) :
							return
						value = self.format_literals(data_type,value)		
						self.assign_value(name,value,scope=scope,scope_name=scope_name)
						return True
					if data_type == 'chars' :
						value = self.format_char_literals(line[2:])
						if not self.check_char_literals(name,value) :
							return
						self.assign_value(name,value,scope=scope,scope_name=scope_name)
						return True
				else :
				 	print("Error from process_variable unit variable is not defined {}".format(line))
				 	return
	
	def output_func(self,line,scope='global',scope_name='') :
		if '(' in line :
			line = line.split()
			if len(line) > 1 :
				return
			l = ""
			for i in line :
				l += i
			line = l
			del l
			expr = ''
			has_found = False
			for i in range(len(line)-1) :
				if line[i] == '(' :
					has_found = True
					continue
				if has_found :
					expr += line[i]
			if expr == '' :
				return
			if expr[0] == "'" and expr[-1] == "'" :
				print(expr)
				return
			expr = expr.split(',')
			if len(expr) >= 1 :
				for v in expr :
					has_op = False
					for op in ['*','/','+','-'] :
						if op in v :
							has_op = True
							break
					if has_op :
						if " " in v :
							e = v.split()
							expr_ = ''
							for i in e :
								expr_ += i
							v = expr_
						tree = ExpressionTree(Variables=self.Variables,scope=scope,scope_name=scope_name)
						tree.buildTree(v)
						print(tree.computeTree(tree.head))
						continue
					for i in v :
						if i in self.alpha :
							if scope == 'global' :
								try :
									val,_ = self.Variables[scope][v]
									print(val,' ')
									break
								except :
									pass
							if scope == 'local' :
								if v in self.Variables['local']['local_scope'][scope_name].keys():
									val,_ = self.Variables[scope]['local_scope'][scope_name][v]
									print(val," ")
									break
								if v in self.Variables['global'].keys() :
									v,_  = self.Variables['global'][v]
									print(v)
									break
						else :
							
							print(v," ")
							break
			return
	
		
	def has_comment(self,line) :
		if self.comment_syntax in line.split()[0] :
			return True
		return False
		
	def length_syntax(self,line) :
		#|array|
		array = line.split()[-1]
		if array[0] == '|' and array[-1] == '|' :
			return True
		return False
	
	#def is_char(self,var) :
		#if var[0] == "'" and var[-1] == "'" :
			#return True
	
	def execute_Functions(self,function,scope='global',scope_name='') :
		name,params,_ = extract_name_param(function)
		
		if name in self.functions.keys() :
			func = self.functions[name]
			func.outer_scope = scope_name
			return func.executeTree(function)
		
		if name in self.func_body.keys() :
			return self.process_functions(function,scope=scope,scope_name=scope_name,exe=True)
			
			
	def to_int(self,n) :
		if type(n) == type('a') :
			return int(n)
		res = []
		for i in n :
			res.append(int(float(i)))
		return res
		
	def numberise_array(self,arr,dim=1) :
		if dim == 1 :
			for i,v in enumerate(arr) :
				try :
					arr[i] = float(v)
				except :
					arr = self.numberise_array(arr,dim=2)
			return arr
		if dim == 2 :
			for i,row in enumerate(arr) :
				for j,col in enumerate(row) :
					arr[i][j] = float(col)
			return arr
			
	def function_to_execute(self,line,scope,scope_name,name) :
		if name == 'condition_split' :
			return condition_split(line,'any')
	
	def length_of_array(self,line,scope='global',scope_name="",get_n=False) :
		#let n = |array|
		#n = |n|
		#arr[i-1] = |n|
		pos = 0
		if get_n :
			variable = ""
			N = len(line)
			
			for i in range(1,N-1) :
				variable += line[i]
				
			elements,_ = self.fetch_variable(variable,scope=scope,scope_name=scope_name)
			elements = self.numberise_array(elements)
			col = None
			
			try :
				col = len(elements[0])
			except :
				pass
			
			if col :
				return len(elements),col
			
			return len(elements)
		
		line = line.split()
		
		
		if len(line) == 4 :
			data_type = line[0]
			pos = 1
			

		variable = line[pos]
		name_bar = line[2+pos]
		name = ''
		
		N = len(name_bar)
		
		for i in range(1,N-1) :
			name += name_bar[i]
		
		elements,_ = self.fetch_variable(name,scope=scope,scope_name=scope_name)
		elements = self.numberise_array(elements)
		n = len(elements)
		col=None
		
		try :
			col = len(elements[0])
		except :
				pass
				

		if pos :
			#print(line)
			if col :
				self.define_variable(variable,[n,col],'int_arr',scope=scope,scope_name=scope_name)
				return
			self.define_variable(variable,n,data_type,scope=scope,scope_name=scope_name)
			return
		if col :
			_,dt = self.fetch_variable(variable,scope=scope,scope_name=scope_name)
			if not dt in ['int_arr','float_arr'] :
				if dt == 'let' :
					dt = 'float_arr'
				else :
					dt = dt+'_arr'
					
			self.define_variable(variable,[n,col],dt,scope=scope,scope_name=scope_name)
			return
			
		self.assign_value(variable,n,scope=scope,scope_name=scope_name)
		
	def initialize_array(self,line,scope='global',scope_name='') :
		
		if type(line) == type('a') :
			line = line.split()
		
		if not self.element in line :
			return False
		if not line[0] in self.ArrayDataTypes :
			return False
			
		def get_non_zero_element(arr) :
			for i in arr :
				if i != 0 :
					return i
			return arr
		
		def zero() :
			return  0
			
		def randint() :
			return random.randint(1,1000)
		
		#int_arr A ∈ R[m,n]
		data_type = line[0]
		variable = line[1]
		self.check_variable_name(variable)
		Set = line[3][0]
		dimension = 1
		size = ''
		dim = line[3][1:]
		N = len(dim)
		m = n = 0
		function = zero
		
		if Set == 'R' :
			function = random.random
		elif Set == 'N' :
			function = randint
		
		for i in range(1,N-1) :
			size += dim[i]
			
		if len(size.split(',')) > 1 :
			dimension = 2
			m,n = size.split(',')
		
		if dimension == 1 :
			m = size
		
		vars = [m,n]

		# m  and will be either number or a variable
		for i,var in enumerate(vars) :
			if var == 0 :
				dimension = 1
				continue
			if self.is_number(var) :
				vars[i] = int(float(var))
			elif self.check_variable_name(var) :
				var = self.fetch_var_value(var,scope=scope,scope_name=scope_name)
				vars[i] = int(float(var))
		
		if dimension == 1 :
			M = None
			if len(vars) == 1 :
				M = vars[0]
			else :
				M = max(vars[0],vars[1])
			arr = [function() for i in range(M)]
		else :
			m,n = vars
			arr = [ [function() for j in range(n)] for i in range(m)]
		
		self.define_variable(variable,arr,data_type,scope=scope,scope_name=scope_name)
		return True
		
	def line_to_execute(self,line,scope,scope_name,to_exe) :
		if to_exe == 'accessing_attributes' :
			return accessing_attributes(None,line,scope=scope,scope_name=scope_name,compiler=self)
	
	def add_to_container(self,c,v) :
		if type(c) == type('a') :
			c += v
		elif type(c) == type([1]) :
			c.append(v)
		return c
	
	def split_space(self,line,to_use='list') :
		new_line = None
		if to_use == 'str' :
			new_line = ''
		else :
			new_line = []
		
		N = len(line)
		found_eq = False
		chars = ''
		for i,l in enumerate(line) :
			if l == '=' :
				found_eq = True
				continue
			if found_eq :
				new_line = self.add_to_container(new_line,'=')
				v = self.split_space(line[i:],to_use='str')
				new_line = self.add_to_container(new_line,v)
				break
			if l != ' ' :
				chars += l
				if i+1 == N :
					v = self.add_to_container(new_line,chars)
					new_line = v
					break
				continue
			new_line = self.add_to_container(new_line,chars)
			chars = ''
		
		return new_line
	
	def ProcessLine(self,line,scope='',scope_name='',class_obj=None,Class=None) :
		if line.split() == [] :
			return
		if self.has_comment(line) :
			return
		elif line.split()[0] == self.break_statement :
			return self.break_statement
		elif line.split()[0] == self.continue_statement :
			return self.continue_statement
		elif self.length_syntax(line) :
			return self.length_of_array(line,scope=scope,scope_name=scope_name)
		elif self.initialize_array(line,scope=scope,scope_name=scope_name) :
			return
		elif accessing_attributes(Class,line,scope=scope,scope_name=scope_name,compiler=self) :
			return
		self.process_variable(line,scope=scope,scope_name=scope_name,class_obj=class_obj)
		self.process_Func(line,scope=scope,scope_name=scope_name)
		self.process_functions(line,scope=scope,scope_name=scope_name)
		self.iostream.ans(line,scope=scope,scope_name=scope_name)
		self.iostream.ask(line,scope=scope,scope_name=scope_name)
		#process_functions(self,line,scope=scope,scope_name=scope_name)
		self.process_array(line,scope=scope,scope_name=scope_name,class_obj=class_obj)	
		process_logic_element(self,line,scope=scope,scope_name=scope_name)
		#accessing_attributes(Class,line,scope=scope,scope_name=scope_name,compiler=self)
	
	def ProcessLines(self,lines) :
		if type(lines) == type('a') :
			if '\n' in lines :
				lines = lines.split('\n')
		count = 0
		i = 0
		loop_count = 0
		N = len(lines)
		while i < N :
			line = lines[i]
			if line.split() == [] :
				i = i + 1
				continue
			if line.split()[0] == 'terminate' :
				return
			if line.split()[0] == 'class' :
				name = line.split()[1]
				self.Variables['local']['local_scope'][name] = {}
				clspu = CLSPU(scope='local',scope_name=name,indent=self.indent,compiler=self)
				pos = clspu.build_class(lines[i:])
				i = i + pos
				continue
			if line.split()[0] == 'Func' :
				name,_,_ = self.get_func_param(line)
				if len(name.split()) > 1 :
					name = name.split()[1]
				self.Variables['local']['local_scope'][name] = {}
				Func = FPU(compiler=self,scope='local',scope_name=name)
				self.functions[name] = Func
				_,pos = Func.buildTree(lines[i:],indent=self.indent)
				i = i + pos
				continue
			if line.split()[0] in self.loop_types :
				self.Variables['local']['local_scope']['for'+str(loop_count)] = {}
				loop = LoopProcessingUnit(self.Variables,scope='local',scope_name='for'+str(loop_count),indent=self.indent,process_variable=self.process_variable,output_func=self.output_func,obj=self)
				pos = loop.buildTree(lines[i:])
				loop.executeTree(node=loop.loop_node)
				loop_count = loop_count + 1
				i = i + pos
				continue
			if line.split()[0]  == 'if' :
				self.Variables['local']['local_scope']['if_0'] =  {}
				cpu = CPU(scope='local',scope_name='if_0',compiler=self)
				cpu.indent = 0
				pos = cpu.buildTree(lines[i:],indent=self.indent)
				cpu.executeTree(cpu.head)
				i  = i + pos
				continue
			self.ProcessLine(line,scope='global')
			i= i + 1
		return
		
	def compile(self,file_name,op="android") :
		self.Variables['local']['local_scope'].clear()
		self.Variables['global'].clear()
		self.functions.clear()
		self.func_body.clear()
		root = ""
		if op == "android" :
			root = '/storage/emulated/0/'
		file = open(root+file_name+'.bhask')
		lines = file.read()
		self.ProcessLines(lines)
		
		


operators = ['+','-','*','/','^']
open_brackets = ['[','{','(']
closing_brackets = [']',')','}']


def get_number(expr,op) :
	chars = ''
	N = len(expr)
	k = 0
	for i in range(N) :
		if not expr[i] in op :
			chars += expr[i]
		else :
			k = i
			break
	return chars,k
	
def trim_space(line) :
	clean_line = ''
	for i in line :
		if i == ' ' :
			continue
		clean_line += i
	return clean_line

def condition_split_v2(expr,op='any') :
	space = ' '
	if space in expr :
		expr = trim_space(expr)
	global operators
	global open_brackets
	global closing_brackets
	found_bracket = False
	elements = []
	element   = ''
	count_ob = 1
	count_cb = 1
	flag  = False
	N = len(expr)
	k  = 0
	if op == 'any' :
		op = operators

	for i in range(N) :
		
		if i+k == N :
			break
		
		if expr[i+k] in open_brackets :
			found_bracket = True
			element += expr[i+k]
			count_ob = count_ob + 1
			continue
		if expr[i+k] in closing_brackets :
			element += expr[i+k]
			count_cb = count_cb + 1
			if count_cb == count_ob :
				found_bracket = False
			if i+k+1 >= N :
				elements.append(element) 
			continue
			 
		if found_bracket :
			element += expr[i+k]
			continue
		
		
		if expr[i+k] == '-' and expr[i+k+1] == '-' :
			# - - = +
			# a - - b = a + b
			a = element
			b,k = get_number(expr[i+k+2:],operators)
			element = a + '+' + b
			k = k + 2
			M = len(element)
			if op == 'any' :
				elements.append(a)
				elements.append(b)
				flag = True
			
			if op == '+' :
				elements.append(a)
				element = b
				flag  = True
				
			
			if M == N - 1 :
				if not flag :
					elements.append(element)
				else :
					elements.append(b)
				break
		
		if expr[i+k] == '-' and expr[i+k+1] == '+' or expr[i+k] == '+' and expr[i+k+1] == '-' :
			# - + = -
			# + - = -
			# a - + b = a - b
			 a  = element
			 b,k = get_number(expr[i+k+2:],operators)
			 element = a + '-' + b
			 k = k + 2
			 M = len(element)
			 
			 if op == 'any' :
			 	elements.append(a)
			 	elements.append(b)
			 	flag = True
			 	
			 
			 if op == '-' :
			 	elements.append(a)
			 	element = b
			 	flag = True
			 
			 if M == N - 1 :
			 	if not flag :
			 		elements.append(element)
			 	else :
			 		elements.append(b)
			 	break
			 
		if expr[i+k] == '-' :
			if i+k-1 == -1 :
				element += '-'
				continue
			
			#*-
			#*/
			if expr[i+k-1] in ['*','/','^'] :
				element += '-'
				continue
		
		if not expr[i+k] in op :
			element += expr[i+k]
			if i+k+1 == N :
				 elements.append(element)
				 break
			continue
		else :
			if flag :
				elements.append(element)
				element = ''
				flag = False
				continue
			elements.append(element)
			element = ''
	return elements

"""
line = '-7--7'
line2 = '-3*-2--3+4'
print(condition_split_v2(line,'+'))
print(condition_split_v2(line2,'any'))
v1= condition_split_v2('-6-+6*8','-')
print(v1)
print(condition_split_v2(v1[0],'-'))
print(condition_split_v2('-3--4*3-9','+'))
print(condition_split_v2('-arr[I+1]--a','+'))
line = '-6--6*8'
print(condition_split_v2(line,'+'))
print(condition_split_v2(' a*fact(x)','*'))
print(condition_split_v2('a*fact(x)','*'))
print(condition_split_v2('2--2','+'))
print(condition_split_v2('6*2+2','+'))
res = condition_split_v2('6*6+3','+')
print(res)
print(condition_split_v2(res[0],'*'))
print(condition_split_v2('2^-2','-'))
"""


def extract_name_param(line) :
	name = ''
	param = ''
	is_found = False
	#f(g(x))
	ob_count = 0
	N = len(line)
	for i in range(N-1) :
		if line[i] == '(' :
			if ob_count == 1 :
				param += line[i]
				continue
			ob_count += 1
			is_found = True
			continue
		if not is_found :
			name += line[i]
			continue
		param += line[i]
	
	return name,param,is_found
	
#print(get_func_param('pow(a,a,b)'))


from .StringAlgorithm.ConditionSplit import condition_split

def process_logic_element(compiler,line,scope='global',scope_name='') :
	
	element = '∈'
	not_element = '∉'
	
	if  element in line :
		if type(line) == type('a') :
			line = line.split()
		
		# a1 ∈ S
		#a[i] ∈ S // this mean i_th element of set S is assigned to variable a
		name = ""
		index = ""
		
		variable = line[0]
		set = line[2]
		has_bracket = False
		
		for char in variable :
		 	 if char == ']' :
		 	 	break
		 	 if char == '[' :
		 	 	has_bracket = True
		 	 	continue
		 	 if has_bracket :
		 	 	index += char
		 	 	continue
		 	 if char in compiler.alpha :
		 	 	name += char
		 	 elif char in compiler.int_literals :
		 	 	index += char
		
		if has_bracket :
			if len(condition_split(index,'any')) > 1 :
				index = compiler.compute_index(index,scope=scope,scope_name=scope_name)
			elif compiler.check_variable_name(index) :
				index = compiler.fetch_var_value(index,scope=scope,scope_name=scope_name)
		
		index = int(index)
		#fetching element at a given index 
		value = compiler.get_arr_element(set,index,scope=scope,scope_name=scope_name)
		#fetching data type of the set
		_,data_type = compiler.fetch_variable(set,scope=scope,scope_name=scope_name)
		
		dt  = ""
		#int_arr
		#int
		for char in data_type :
			if char == '_' :
				break
			dt += char
		data_type = dt
		del dt
		
		if compiler.is_variable_in_scope(name,scope=scope,scope_name=scope_name) :
			compiler.assign_value(name,value,scope=scope,scope_name=scope_name)
			return
		
		compiler.define_variable(name,value,data_type,scope=scope,scope_name=scope_name)
	
	return



open_brackets = ['[','{','(']
closing_brackets = [']','}',')']

def trim_space(line) :
	new_line = ''
	space = ' '
	for char in line :
		if char == space :
			continue
		new_line += char
	return new_line
	
def get_num(line,op,k,N) :
	b = ''
	M = len(line)
	for i in range(0,M) :
		if not line[i] in op :
			b += line[i]
			continue
		k = k+i+1
		break
	return b,k

def condition_split_v1(expr,oper='any') :
	elements = []
	element = ''
	operators = ['-','+','/','*','^']
	global open_brackets
	global closing_brackets
	op = operators
	if op != 'any' :
		op = oper
	i = 0
	found_bracket = False
	N = len(expr)
	count_ob = 0
	seperate = False
	k = 0
	count_cb = 0
	for i in range(N) :
		
		if i+k >= N :
			break
			
		if expr[i+k] in open_brackets :
			element += expr[i+k]
			count_ob += 1
			found_bracket = True
			continue
		if expr[i+k] in closing_brackets :
			element += expr[i+k]
			count_cb += 1
			if count_cb == count_ob :
				found_bracket = False
				elements.append(element)
				element = ""
			continue
		if found_bracket :
			element += expr[i+k]
			continue
		
		if expr[i+k] == '-' and expr[i+k+1] == '-' :
			#  -  -  = +
			# a - - b = a + b
			# a - - b*3 = a + b * 3+1
			a = element
			b,k = get_num(expr[i+k+2:],operators,i+k+2,N)
			element = a + '+' + b
			
			if oper == '+' :
				elements.append(a)
				elements.append(b)
				seperate = True
				continue
			
			if oper == 'any' :
				elements.append(a)
				elements.append(b)
				seperate = True
				continue
			if i+k+ 1 >= N :
				elements.append(element)
			continue

		if expr[i+k] == '-' and expr[i+k+1] == '+' or expr[i+k] == '+' and expr[i+k+1] == '-' :
			# - + = -
			# + - = -
			# -a - + b  = -a  - b
			# a + - b = a - b
			a = element
			b,k = get_num(expr[i+k+2:],operators,i+k+2,N)
			element = a + '-' + b
			
			if oper == '-' :
				elements.append(a)
				elements.append(b)
				seperate = True
				continue
			
			if oper == 'any' :
				elements.append(a)
				elements.append(b)
				seperate = True
				continue
			
			if i+k+1 >= N :
				elements.append(element)
			continue
		
		if expr[i+k] == '-' :
			# *-
			# /-
			
			if i+k-1 == -1 :
				element += '-'
				continue
			elif expr[i+k-1] in ['*','/','^'] :
				if oper == 'any' :
					elements.append(element)
					element = '-'
					continue
				element += '-'
				continue
		
		if not expr[i+k] in op :
			element += expr[i+k]
			if i+k + 1 == N :
				elements.append(element)
				break
			continue
		else :
			if seperate :
				element = ""
				seperate = False
				continue
			elements.append(element)
			element = ""
	return elements

"""
line = '-2*-3*-4+7--7'
line1 = '-7+-7'
print(condition_split_v1(line,'any'))
print(condition_split_v1(line1,'-'))
print(condition_split_v1('-2+3','+'))
print(condition_split_v1('8*3--3','-'))
print(condition_split_v1('2^-2','-'))
"""


class LoopProcessingUnit:
	
	def __init__(self,indent="    ") :
		self.indent = indent
		if len(self.indent) == 0 :
			self.indent = "    "
		self.allowed_loops = ['for','while']
	
	def has_for_loop_syntax(self,line) :
		if line[2] == '=' :
			if line[4] == 'to' :
				if line[6] == ':' :
					return True
				return "Forget indentation [:]"
			return "Forgot [to]"
		return "Forgot [=]"
	
	def check_variable(self,var) :
		nums = '0123456789'
		alpha = 'abcdefghijklmnopqrstuvwxyz'
		if len(var) > 1 :
			if var[0] in nums :
				for v in range(1,len(var)) :
					if not var[v] in nums :
						print("{} incorrect variable name ".format(var))
						return False
				return True
			if var[0] in alpha :
				for v in range(1,len(var)) :
					if not var[v] in nums and not var[v] in alpha :
						print("{} incorrect variable name".format(var))
						return False
				return True
		if var in alpha :
			return True
		if var in nums :
			return True
		return False 
	
	def has_while_loop_syntax(self,line) :
		#while i [<,>,<=,>=] a :
		comparsion_operator = ['>','<','<=','>=']
		if line[2] in comparsion_operator :
			if line[-1] == ":" :
				return True
			return "Forgot indentation :"
		return "incorrect operator used in while loop"
	
	def check_syntax(self,line) :
		#for i = a to b :
		#while a [<,>,<=,>=] n :
		if type(line) == type([1]) :
			pass
		else :
			line = line.split()
		if line == [] :
			return False
		if len(line) == 7 :
			if line[0] == 'for' :
				val = self.has_for_loop_syntax(line)
				if val == True :
					i = line[1]
					a = line[3]
					b = line[5]
					vars = [i,a,b]
					bool_list = [self.check_variable(var) for var in vars]
					if False in bool_list :
						print("Variables are defined incorrectly in loop header")
						return False
					return True
				print('\n',val)
				return -2
		if len(line) == 5 :
			if line[0] == "while" :
				var = self.has_while_loop_syntax(line)
				if var == True :
					# while a < b :
					a = line[1]
					b = line[3]
					vars = [a,b]
					bool_list = [self.check_variable(var) for var in vars]
					if False in bool_list :
						print("Variables are defined incorrectly in loop header")
						return -2
					return True
				print('\n',var)
				return -2
		return False
		
	def has_valid_indent(self,line,bias_indent=0) :
		count = 0
		for l in line :
			if l != ' ' :
				if count != len(self.indent)+bias_indent :
					return False
				else :
					break
			if l == ' ' :
				if count == len(self.indent)+bias_indent :
					if self.has_valid_indent(line.split('    ')) :
						continue
					return False
				count += 1
		return True
		
	def Fetch_Loop_Body(self,lines,start,bias_indent=0) :
		loop_body = []
		line_num = 0
		print(lines)
		if type(lines) == type('a') :
			lines = lines.split('\n')
		for ii in range(start,len(lines)) :
			if self.has_valid_indent(lines[ii],bias_indent) :
				if lines[ii].split() == [] :
					continue
				loop_body.append(lines[ii].split())
			else :
				line_num = ii
				break
		return loop_body,line_num
	
	def Is_Num(self,st) :
		nums = '0123456789'
		for s in st :
			if not s in nums :
				return False
		return True
	
	def EvaluateLoop(self,loop_body) :
		loop_syntax = loop_body[0]
		if loop_syntax[0] == 'for' :
			#for i = a to b :
			step = 1
			i = loop_syntax[1]
			a = loop_syntax[3]
			b = loop_syntax[5]
			try :
				a = int(a)
				b = int(b)
			except :
				pass
			if a > b :
				step = -1
			for ii in range(a,b+1,step) :
				print(ii)
				#and also evaluating all loop bodies excpression
		if loop_syntax[0] == 'while' :
			#while a < b
			step = 1
			a = int(loop_syntax[1])
			b = int(loop_syntax[3])
			op = loop_syntax[2]
			if op == '<' :
				for i in range(a,b,step) :
					print(i)
				return
			if op == '>' :
				step = -1
				for ii in range(a,b,step) :
					print(ii)
				return
			if op == '>=' :
				step = -1
				for ii in range(a,b-1,step) :
					print(ii)
				return
			if op == '<=' :
				step = 1
				for ii in range(a,b+1,step) :
					print(ii)
				return
		return
	
	def HasNestedLoop(self,data) :
		print(data)
		for d in range(len(data)) :
			if self.check_syntax(data[d]) :
				return d	
		return False
		
	def process_loop(self,data,st=0,nes=0) :
		if type(data) == type([1]) :
			pass
		else :
			data = data.split('\n')
		loop_body = []
		loop_bodies = {}
		nested_loops = []
		N = len(data)
		ii = st
		count = 0
		while ii < N :
			prev_i = ii
			if self.check_syntax(data[ii]) :
				loop_body,ln = self.Fetch_Loop_Body(data[ii+1:],ii+1,nes)
				nes_ln = self.HasNestedLoop(loop_body)
				print(loop_body)
				if nes_ln :
					nested_loops = self.process_loop(loop_body,nes_ln,nes=nes+4)
				loop_body.insert(0,data[ii].split())
				if nested_loops :
					loop_bodies[str(count)+'_nes'] = nested_loops
				loop_bodies[str(count)] = loop_body
				loop_body = []
				nested_loops = []
				ii = ii + ln
				count = count + 1
			if ii > prev_i :
				continue
			ii = ii + 1	
				
		if loop_bodies == {} :
			return
		return loop_bodies
	
	def ProcessLoop(self,data) :
		loop_bodies = self.process_loop(data)
		for key in loop_bodies.keys():
			self.EvaluateLoop(loop_bodies[key])
		return
		
			
#LPU = LoopProcessingUnit()
Loop = """
for i = 1 to 10 :
     for j = 1 to 10 :
		h = a + 1
	a = b + 1
while 10 > 1 :
	a = n + 2*n	
"""

"""
LPU.ProcessLoop(Loop)
print(LPU.process_loop("for i = 1 to 3 :\n    for j = 1 to 3:\n    "))
print(LPU.check_syntax("for i = 1 to 10 :"))
"""


from .StringAlgorithm.ConditionSplit import condition_split
from .StringAlgorithm.ConditionSplitV3 import condition_split_v2

class Node :
	
	def __init__(self,sym) :
		self.sym = sym
		self.child_nodes = []
		self.expression = []

class ExpressionTree :
	
	def __init__(self,Variables={},scope='',scope_name='',is_func=False,parameters=[],param_values={},obj=None) :
		self.head = None
		self.obj = obj
		self.Variables = Variables
		self.scope = scope
		self.param_values = param_values
		self.is_func = is_func
		self.parameters = parameters
		self.scope_name = scope_name
		self.alpha = 'abcdefghijklmnopqrstuvwxyz'
		self.chars = 'abcdefghijklmnopqrstuvwxyz0123456789'
		self.nums = '0123456789'
		
		
	def is_number(self,num) :
			try :
				a = float(num)
				return True
			except :
				return False
				
	
	def fetchValue(self,n) :
		if self.is_func :
			if n in self.parameters :
				return n
		if self.scope == 'local' :
			if n in self.Variables['local']['local_scope'][self.scope_name].keys() :
				n,_ = self.Variables['local']['local_scope'][self.scope_name][n]
				return n
			else :
				try :
					if n in self.Variables['global'].keys() :
						n,_= self.Variables['global'][n]
						return n
				except :
					print("Keyerror")
		if self.scope == 'global' :
			try :
				n,_= self.Variables['global'][n]
				return n
			except :
				v_ = self.Variables['local']['local_scope'][self.scope_name]
				if n in v_.keys() :
					return v_[n][0]
				print("error {} is not defined ".format(n))
	
	def buildTree(self,line,prev_node=None) :
		
		if len(condition_split_v2(line,'+')) > 1 :
			line = condition_split_v2(line,op='+')
			node = Node('+')
			
			if prev_node :
				prev_node.expression.append(node)
			else :
				self.head = node
			for l in line :
				n = self.buildTree(l,prev_node=node)
				if n :
					if n in self.obj.mathematical_constants.keys() :
						node.expression.append(self.obj.mathematical_constants[n])
						continue
					
					if self.obj.get_arr_index(n) and not '.' in n:
						n,_ = self.obj.process_array(n,scope=self.scope,scope_name=self.scope_name)
						node.expression.append(float(n))
						continue
					
					is_var = False
					
					if self.obj.check_variable_name(n) :
						is_var = True
					
					if is_var :
						n = self.obj.fetch_var_value(n,scope=self.scope,scope_name=self.scope_name)
						node.expression.append(float(n))
						continue
					
					if self.is_number(n) :
						node.expression.append(float(n))
						continue
					
					name,param,is_found = self.obj.get_func_param(n)
					
					if is_found :
						
						if name in self.obj.func_body.keys() :
							n = self.obj.process_functions(n,scope=self.scope,scope_name=self.scope_name,exe=True)
							node.expression.append(float(n))
							continue
							
						if name in self.obj.functions.keys() :
							func = self.obj.functions[name]
							func.outer_scope = self.scope_name
							n = func.executeTree(n)
							
							if self.is_number(n) :
								node.expression.append(float(n))
								continue
							
							if type(n) == type('a') :
								node.expression.append(float(n))
								continue
							
							if type(n) == type(['1']) :
								for ii in n :
									node.expression.append(float(ii))
								continue
					if not self.is_number(n) and '.' in n :
						node.expression.append(float(self.obj.line_to_execute(n,self.scope,self.scope_name,'accessing_attributes')))
						continue
					
					 
					if '-' in n :
						if self.obj.check_variable_name(n[1:]) :
							n = self.obj.fetch_var_value(n[1:],scope=self.scope,scope_name=self.scope_name)
							n = '-' + str(n)
							n = self.minus_minus_plus(n)
							node.expression.append(float(n))
									
			return
		
		if len(condition_split_v2(line,'-')) > 1 :
			line = condition_split_v2(line,op='-')
			node = Node('-')
			if prev_node :
				prev_node.expression.append(node)
			else :
				self.head = node
					
			for l in line :
				n = self.buildTree(l,prev_node=node)
				
				if n :
					if n in self.obj.mathematical_constants.keys() :
						node.expression.append(self.obj.mathematical_constants[n])
						continue
					
					if self.obj.get_arr_index(n) and not '.' in n:
						n,_ = self.obj.process_array(n,scope=self.scope,scope_name=self.scope_name)
						node.expression.append(float(n))
						continue
					
					is_var = False
					
					if self.obj.check_variable_name(n) :
						is_var = True
					
					if is_var :
					    n = self.obj.fetch_var_value(n,scope=self.scope,scope_name=self.scope_name)
					    node.expression.append(float(n))
					    continue
					
					if self.is_number(n) :
						node.expression.append(float(n))
						continue
						
					name,param,is_found = self.obj.get_func_param(n)
					
					if is_found :
					    if name in self.obj.func_body.keys() :
					         n = self.obj.process_functions(n,scope=self.scope,scope_name=self.scope_name,exe=True)
					         node.expression.append(float(n))
					         continue
					    if name in self.obj.functions.keys() :
					        func = self.obj.functions[name]
					        func.outer_scope = self.scope_name
					        n = func.executeTree(n)
					          
					        if self.is_number(n):
					          	node.expression.append(float(n))
					          	continue
					        if type(n) == type('a') :
					            node.expression.append(float(n))
					            continue
					        if type(n) == type(['1']) :
					             for ii in n :
					                 node.expression.append(float(ii))
					             continue
					if not self.is_number(n) and '.' in n :
					    node.expression.append(float(self.obj.line_to_execute(n,self.scope,self.scope_name,'accessing_attributes')))
					    continue   
					 
					if '-' in n :
						if self.obj.check_variable_name(n[1:]) :
							n = self.obj.fetch_var_value(n[1:],scope=self.scope,scope_name=self.scope_name)
							n = '-' + str(n)
							n = self.minus_minus_plus(n)
							node.expression.append(float(n))
					                 
			return
		if  len(condition_split_v2(line,'/')) > 1 :
			line = condition_split_v2(line,'/')
			node = Node('/')

			if prev_node :
				prev_node.expression.append(node)
			else :
				self.head = node
			for l in line :
				n = self.buildTree(l,prev_node=node)
				if n :
					
					if n in self.obj.mathematical_constants.keys() :
						node.expression.append(self.obj.mathematical_constants[n])
						continue
					
					if self.obj.get_arr_index(n) and not '.' in n:
						n,_ = self.obj.process_array(n,scope=self.scope,scope_name=self.scope_name)
						node.expression.append(float(n))
						continue
						
					is_var = False
					
					if self.obj.check_variable_name(n) :
						is_var = True
					
					if is_var :
					    n = self.obj.fetch_var_value(n,scope=self.scope,scope_name=self.scope_name)
					    node.expression.append(float(n))
					    continue
					
					if self.is_number(n) :
					    node.expression.append(float(n))
					    continue
					    	
					name,param,is_found = self.obj.get_func_param(n)
					if is_found :
					    if name in self.obj.func_body.keys() :
					    	n = self.obj.process_functions(n,scope=self.scope,scope_name=self.scope_name,exe=True)
					    	node.expression.append(float(n))
					    	continue
					    if name in self.obj.functions.keys() :
					    	func = self.obj.functions[name]
					    	func.outer_scope = self.scope_name
					    	n = func.executeTree(n)
					    	
					    	if self.is_number(n) :
					    		node.expression.append(float(n))
					    		continue
					    	if type(n) == type('a') :
					    		node.expression.append(float(n))
					    		continue
					    	if type(n) == type(['1']) :
					    		for ii in n :
					    			node.expression.append(float(ii))
					    		continue
					if not self.is_number(n) and '.' in n :
						value = self.obj.line_to_execute(n,self.scope,self.scope_name,'accessing_attributes')
						node.expression.append(float(value))
						continue
					
					 
					if '-' in n :
						if self.obj.check_variable_name(n[1:]) :
							n = self.obj.fetch_var_value(n[1:],scope=self.scope,scope_name=self.scope_name)
							n = '-' + str(n)
							n = self.minus_minus_plus(n)
							node.expression.append(float(n))
			return
		
		if len(condition_split_v2(line,'*')) > 1 :
			line = condition_split_v2(line,op='*')
			node = Node('*')
			
			if prev_node :
				prev_node.expression.append(node)
			else :
				self.head = node
				
			for l in line :
				n = self.buildTree(l,prev_node=node)
				
				if n :
					
					if n in self.obj.mathematical_constants.keys() :
						node.expression.append(self.obj.mathematical_constants[n])
						continue
						
					
					if self.obj.get_arr_index(n) and not '.' in n:
						n,_ = self.obj.process_array(n,scope=self.scope,scope_name=self.scope_name)
						node.expression.append(float(n))
						continue
						
					is_var = False
					if self.obj.check_variable_name(n) :
						is_var = True
						
					if is_var :
					    n = self.obj.fetch_var_value(n,scope=self.scope,scope_name=self.scope_name)
					    if n :
					    	node.expression.append(float(n))
					    continue
					
					if self.is_number(n) :
						node.expression.append(float(n))
						continue
					
					name,param,is_found = self.obj.get_func_param(n)
					if is_found :
					    name ,param,_ = self.obj.get_func_param(n)
					    if name in self.obj.func_body.keys() :
					    	n = self.obj.process_functions(n,scope=self.scope,scope_name=self.scope_name,exe=True)
					    	node.expression.append(float(n))
					    	continue
					    if name in self.obj.functions.keys() :
					    	func = self.obj.functions[name]
					    	func.outer_scope = self.scope_name
					    	n = func.executeTree(n)
					    	if self.is_number(n) :
					    		node.expression.append(float(n))
					    		continue
					    	if type(n) == type('a') :
					    		node.expression.append(float(n))
					    		continue
					    	if type(n) == type(['1']) :
					    		for ii in n :
					    			node.expression.append(float(ii))
					    		continue
					    		
					if not self.is_number(n) and '.' in n :
					    node.expression.append(float(self.obj.line_to_execute(n,self.scope,self.scope_name,'accessing_attributes')))
					    continue
					
					 
					if '-' in n :
						if self.obj.check_variable_name(n[1:]) :
							n = self.obj.fetch_var_value(n[1:],scope=self.scope,scope_name=self.scope_name)
							n = '-' + str(n)
							n = self.minus_minus_plus(n)
							node.expression.append(float(n))
			return
			
		if len(condition_split_v2(line,'^')) > 1 :
			line = condition_split_v2(line,op='^')
			node = Node('^')
			
			if prev_node :
				prev_node.expression.append(node)
			else :
				self.head = node
				
			for l in line :
				n = self.buildTree(l,prev_node=node)
				
				if n :
					
					if n in self.obj.mathematical_constants.keys() :
						node.expression.append(self.obj.mathematical_constants[n])
						continue
						
					
					if self.obj.get_arr_index(n) and not '.' in n:
						n,_ = self.obj.process_array(n,scope=self.scope,scope_name=self.scope_name)
						node.expression.append(float(n))
						continue
						
					is_var = False
					if self.obj.check_variable_name(n) :
						is_var = True
						
					if is_var :
					    n = self.obj.fetch_var_value(n,scope=self.scope,scope_name=self.scope_name)
					    if n :
					    	node.expression.append(float(n))
					    continue
					
					if self.is_number(n) :
						node.expression.append(float(n))
						continue
					
					name,param,is_found = self.obj.get_func_param(n)
					if is_found :
					    name ,param,_ = self.obj.get_func_param(n)
					    if name in self.obj.func_body.keys() :
					    	n = self.obj.process_functions(n,scope=self.scope,scope_name=self.scope_name,exe=True)
					    	node.expression.append(float(n))
					    	continue
					    if name in self.obj.functions.keys() :
					    	func = self.obj.functions[name]
					    	func.outer_scope = self.scope_name
					    	n = func.executeTree(n)
					    	if self.is_number(n) :
					    		node.expression.append(float(n))
					    		continue
					    	if type(n) == type('a') :
					    		node.expression.append(float(n))
					    		continue
					    	if type(n) == type(['1']) :
					    		for ii in n :
					    			node.expression.append(float(ii))
					    		continue
					    		
					if not self.is_number(n) and '.' in n :
					    node.expression.append(float(self.obj.line_to_execute(n,self.scope,self.scope_name,'accessing_attributes')))
					    continue
					    
					if '-' in n :
						if self.obj.check_variable_name(n[1:]) :
							n = self.obj.fetch_var_value(n[1:],scope=self.scope,scope_name=self.scope_name)
							n = '-' + str(n)
							n = self.minus_minus_plus(n)
							node.expression.append(float(n))
			return
			
			
		return line
	
	def minus_minus_plus(self,num) :
		if len(num) >=3 :
			if num[0] == '-' and num[1] == '-' :
				return num[2:]
		return num
	
	def list_cal(self,list_,symb) :
		
		if symb == '+' :
			res = 0
			for l in list_ :
				res = res + l
			return res
		
		if symb == '-' :
			res = list_[0]
			for i in range(1,len(list_)) :
				res = res - list_[i]
			return res
		
		if symb == '*' :
			res  = 1
			for l in list_ :
				res = res * l
			return res
		
		if symb == '/' :
			res = list_[0]
			for i in range(1,len(list_)) :
				try :
					res = res/list_[i]
				except :
					res = 0
			return res
			
		if symb == '^' :
			return list_[0]**list_[1]
		
	def computeTree(self,node) :
		if node :
			res = 0
			ops = []
			if node.expression :
				for child in range(len(node.expression)) :
					if type(node.expression[child]) == type(node) :
						N = self.computeTree(node.expression[child])
						ops.append(N)
						continue
					if node.expression[child] in self.parameters :
						ops.append(float(self.param_values[node.expression[child]]))
						continue
					ops.append(node.expression[child])
				return self.list_cal(ops,node.sym)
				
					
	
	def printTree(self,node) :
		if node :
			if node.expression :
				for child in range(len(node.expression)) :
					if type(node.expression[child]) == type(node) :
						self.printTree(node.expression[child])
						continue
					if node.expression[child] in self.parameters :
						node.expression[child] = float(self.param_values[node.expression[child]])
					print(node.expression[child],node.sym," ",end="")
				return
			return
		return
	"""
	def printTree(self,node) :
		if node :
			if node.child_nodes :
				for child in range(len(node.child_nodes)):
						self.printTree(node.child_nodes[child])
						if child + 1 == len(node.child_nodes) :
							continue
						print(node.sym,end='')
						
				print(node.sym," ",end='')
			if node.expression :
				for ii in range(len(node.expression)) :
					if ii + 1 < len(node.expression) :
						print(node.expression[ii],' ',node.sym,' ',end='')
						continue
					print(node.expression[ii],' ',end=' ')
				return
			return
		return
		"""
		
		
"""
expr_tree = ExpressionTree()
expr = '8*8 + 2 + 3- 7*7 +8-8'
expr_tree.buildTree(expr)
print(expr)
expr_tree.printTree(expr_tree.head)
print('\n')
print(expr_tree.computeTree(expr_tree.head))
"""
"""
a = "2*2 + 3*2"
tree = ExpressionTree()
tree.buildTree(a)
print(tree.computeTree(tree.head))
"""


from setuptools import setup, find_packages

setup(
    name="Bhaskara",  # Replace with your package name
    version="4.1.6.4.6",
    description="Bhaskara Programming Engine",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    #url="https://github.com/username/my_package",  # Replace with your repository
    author="Prakamya Khare",
    author_email="khareprakamya27@gmail.com",
    license="MIT",  # Or other license you use
    packages=find_packages(),  # This will include your package
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[],  # List any dependencies here
)



class Node :
	
	def __init__(self,op) :
		self.op = op
		self.child_nodes = []
		self.vars = []


class ExpressionTree() :
	
	def __init__(self,) :
		self.head = None
		self.ops = ['-','+','/','*']
		self.nums = '0123456789'
	
	def Compute_Scalers(self,symb,a,b) :
		if symb == '+' :
			return a + b
		if symb == '-' :
			return a  - b
		if symb == '/' :
			return a / b 
		if symb == '*' :
			return  a * b
	
	def Compute_List(self,symb,num_list) :
		if len(num_list) == 1 :
			return num_list[0]
		if symb == '+' :
			res = 0
			for n in num_list :
				res = res + n
			return res
		if symb == '-' :
			res = num_list[0]
			for i in range(1,len(num_list)) :
				res = res - num_list[i]
			return res
		if symb == '*' :
			res = 1
			for n in num_list :
				res = res * n
			return res
		if symb == '/' :
			res = num_list[0]
			for n in range(1,len(num_list)) :
				res = res / num_list[n]
			return res
	
	def strip_white_space(self,expr) :
		stripped_expr = ''
		for exp in expr :
			if exp == ' ' :
				continue
			stripped_expr += exp
		return stripped_expr
	
	def buildTree(self,expr,prev_node=None) :
		for op in self.ops :
			if op in expr :
				expr = expr.split(op)
				node = Node(op)
				if prev_node :
					prev_node.child_nodes.append(node)
				else :
					self.head = node
				for exp in expr :
					exp = self.strip_white_space(exp)
					if exp in self.nums :
						node.vars.append(exp)
					else :
						self.buildTree(exp,prev_node=node)
		return
	
	def ComputeExpression(self,node=None) :
		if node :
			if node.child_nodes :
				res = []
				var_res = None
				for child in node.child_nodes :
					temp = self.ComputeExpression(node=child)
					res.append(temp)
				res_total = self.Compute_List(node.op,res)
				if node.vars :
					casted_vars = []
					for v in node.vars :
						casted_vars.append(float(v))
					var_res = self.Compute_List(node.op,casted_vars)
					if not res_total :
						return var_res
				if not var_res :
					return res_total
				return self.Compute_Scalers(node.op,res_total,var_res)
			if node.vars :
				casted_vars = []
				for v in node.vars :
					casted_vars.append(float(v))
				var_res = self.Compute_List(node.op,casted_vars)
				return var_res
	
	def printTree(self,node=None) :
		if node :
			if node.child_nodes :
				for child in range(len(node.child_nodes)) :
					self.printTree(node=node.child_nodes[child])
					if child+1 == len(node.child_nodes) :
						continue
					print(node.op,' ',end='')
			if node.vars :
				for v in range(len(node.vars)) :
					print(node.op,node.vars[v],' ',end='')
		return			
							
"""
tree = ExpressionTree()
expr = "2*8 + 7 - 6 * 4/5"
tree.buildTree(expr)
tree.printTree(tree.head)
print('\n')
print(tree.ComputeExpression(tree.head))
"""


from .StringAlgorithm.ConditionSplit import condition_split
from .ExpressionTree1 import ExpressionTree as EXPT

class LoopNode :
	
	def __init__(self,) :
		self.syntax = ""
		self.loop_body = []

class LoopProcessingUnit() :
	
	def __init__(self,Variables=None,scope='local',scope_name='',indent=4,process_variable=None,output_func=None,obj=None,is_func=False):
		self.loop_node = None
		self.indent = indent
		self.scope = scope
		self.is_func = is_func
		self.cpu = None
		self.process_variable = process_variable
		self.output_func = output_func
		self.scope_name = scope_name
		self.obj = obj
		self.Variables = Variables
		self.loop_types = ['for','while']
		self.vars = Variables['local']['local_scope'][self.scope_name]
		self.alpha = 'abcdefghijklmnopqrstuvwxyz'
	
	def hasIndent(self,line,indent=0) :
		count = 0
		for l in line :
			if l == " " :
				count += 1
				continue
			break
		if count == self.indent+indent :
			return True
		return False
	
	def is_alpha(self,name) :
		for n in name :
			if n in self.alpha :
				return True
		return False
	
	def process_loop_syntax(self,line) :
		if line == "" or line == [] :
			return
		temp = line
		if type(line) == type('a') :
			temp = line.split()
		if temp[0] == 'for' :
			#for i = 1 to 3 :
			if len(temp) != 7 :
				print("Incorrect loop Syntax {}".format(line))
				return
			if temp[2] != '=' :
				print("Incorrect loop Syntax {}".format(line))
				return
			variable_i = temp[1]
			variable_a = temp[3]
			variable_b = temp[5]
			vars = [variable_a,variable_b]
			if not self.is_alpha(variable_i) :
				print("Error this [i] = 1 must be a variable {}".format(variable_i))
				return
			for i,v in enumerate(vars) :
				
				if self.obj.length_syntax(v) :
					vars[i] = self.obj.length_of_array(v,scope=self.scope,scope_name=self.scope_name,get_n=True)
					continue
				
				if self.obj.is_number(v) :
					vars[i] = int(float(v))
					continue
				
				if len(condition_split(v,'any')) > 1 :
					vars[i] = int(self.obj.compute_index(v,scope=self.scope,scope_name=self.scope_name))
					continue
					
				if self.obj.check_variable_name(v) :
					val,_= self.obj.fetch_variable(v,scope='local',scope_name=self.scope_name)
					vars[i] = int(float(val))
					
			self.vars[variable_i] = [variable_a,'int']
		return vars[0],vars[1]
	 
	def buildTree(self,lines,prev_node=None,indent=0) :
		node = LoopNode()
		node.syntax = lines[0]
		i = 1
		N = len(lines)
		
		if prev_node == None :
			self.loop_node = node
		else :
			prev_node.loop_body.append(node)
			
		while i < N :
			line = lines[i]
			if line.split() == [] :
				i = i + 1
				continue
			if not self.hasIndent(line,indent) :
				break
			if line.split()[0] in self.loop_types :
				pos = self.buildTree(lines[i:],prev_node=node,indent=indent+self.obj.indent)
				i = i + pos
				continue
			if line.split()[0] == 'if' :
				cpu = ConditionalProcessingUnit(scope='local',scope_name=self.scope_name,compiler=self.obj)
				cpu.has_loop = True
				pos = cpu.buildTree(lines[i:],indent=indent+self.indent)
				i = i + pos
				node.loop_body.append(cpu)
				if not self.cpu :
					self.cpu = cpu
				continue
			node.loop_body.append(line)
			i = i + 1
		return i
		
	def executeTree(self,node=None,d=0) :
		if node :
			a,b = self.process_loop_syntax(node.syntax)
			var_a = node.syntax.split()[1]
			for i in range(a,b+1) :
				_,dt = self.vars[var_a]
				self.vars[var_a] = [str(i),dt]
				for line in node.loop_body :
					if type(line) == type(node) :
						return_val = self.executeTree(line,d=d+1)
						if return_val :
							if return_val.split()[0] == self.obj.break_statement :
									return
							if return_val.split()[0] == self.obj.continue_statement :
									break
						continue

					if type(line) == type(self.cpu) :
						return_value = line.executeTree(line.head)
						if return_value :
							arr = return_value.split()
							if arr[0] == 'return' :
								if len(arr) > 1 :
									value = arr[1]
									expr = EXPT()
									expr.buildTree(value)
									return expr.computeTree(expr.head)
								return
						
						if return_value == self.obj.break_statement :
							return
						elif return_value == self.obj.continue_statement :
							break
						continue
					return_val =self.obj.ProcessLine(line,scope='local',scope_name=self.scope_name)
					if return_val == self.obj.break_statement :
						return
					if return_val == self.obj.continue_statement :
						break
			return
from .ConditonalProcessingUnit import ConditionalProcessingUnit


def HasIndent(line,indent) :
	count = 0
	for l in line :
		if l == ' ' :
			count += 1
			continue
		break
	if count == indent :
		return True
	return False

def LoopBody(lines,indent) :
	if type(lines) == type('a') :
		lines = lines.split('\n')
	loop_body = []
	loops = ['for','while']
	i = 0
	N = len(lines)
	while i < N :
		line = lines[i]
		nested_loop = []
		if line.split() == [] :
			i = i + 1
			continue
		if HasIndent(line,indent) :
			if line.split()[0] in loops :
				nested_loops,pos = LoopBody(lines[i+1:],indent+4)
				loop_body.append(line)
				loop_body.append(nested_loops)
				pos = pos + 1
				i = i + pos
				continue
			loop_body.append(line)
			i = i + 1
			continue
		break
	return loop_body,i

def ProcessLoop(lines,indent) :
	if type(lines) == type('a') :
		lines = lines.split('\n')
	loop_bodies = {}
	Loops = ['for','while']
	count = 0
	loop_body = []
	i = 0
	N = len(lines)
	while i < N :
		line = lines[i]
		if line.split() == [] :
			i = i + 1
			continue
		if line.split()[0] in Loops :
			loop_body,pos = LoopBody(lines[i+1:],indent)
			loop_body.insert(0,line)
			loop_bodies[str(count)] = loop_body
			count = count + 1
			pos = pos + 1
			i = i + pos
			continue
		i = i + 1
	return loop_bodies
L = """
for i = 1 to N :
    h = h + 1
    while a < 1 :
        j = j + 1
        for j = 1 to H :
            pass
            pass
            pass
    return True
for i = 1 to N :
    for j = 1 to M :
        for k = 1 to O :
            a  = arr[i,j,k]
    a = 1
h = h + 1
a  = a + 1
"""

l = ProcessLoop(L,4)
print(l)